/* $Id: main.c,v 1.12 2002/11/08 00:28:33 julliard Exp $ */

/*
Modifications by Borland/Inprise Corp.:
    05/17/2000: Added support for two new command-line arguments: 
                -i path: used to specify the install directory. 
		-b path: used to specify the binary directory.
		If either of these parameters are specified on the command
		line, the user will not be asked for them. Two global flags
		were added for this purpose: disable_install_path and 
		disable_binary_path. Additional changes were made to install.c
		gtk_ui.c and console_ui.c were made to support this change.

   06/02/2000:  Added support for a new command-line parameter:
                -m: used to force RPM packages to be manually extracted and to
		    prevent the RPM database from being updated.
		Some distributions (Corel) include RPM support, but the RPM 
		database is empty. This causes almost every attempt to install
		an RPM to fail because of missing dependencies. Using the -m
		parameter allows users to get around this by forcing setup to
		work as if RPM was not present.

 */
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdarg.h>

#include "install_log.h"
#include "install_ui.h"
#include "log.h"
#include "file.h"
#include "detect.h"
#include "plugins.h"

#define SETUP_CONFIG  SETUP_BASE "setup.xml"

#define PACKAGE "setup"
#define LOCALEDIR SETUP_BASE "locale"

/* Global options */

int force_console = 0;
int disable_install_path = 0;
int disable_binary_path = 0;
#ifdef RPM_SUPPORT
char *rpm_root = "/";
int force_manual = 0;
#endif
const char *argv0 = NULL;

static install_info *info = NULL;
Install_UI UI;

/* List of options enabled on the command line */
static struct enabled_option {
    char *option;
    struct enabled_option *next;
} *enabled_options = NULL;

static char *current_locale = NULL;

void exit_setup(int ret)
{
    /* Cleanup afterwards */
    if ( info )
        delete_install(info);
    FreePlugins();
    exit(ret);
}

void signal_abort(int sig)
{
    char manifest[PATH_MAX];
    struct stat dummy;

    signal(SIGINT, SIG_IGN);
    if ( UI.abort )
        UI.abort(info);

    /* Set SETUP_ABORT so scripts know the install is being aborted */
    putenv("SETUP_ABORT=1");
    sprintf(manifest, "%s/.manifest/%s.xml", info->install_path, GetProductName(info));
    if (stat(manifest, &dummy)<0)
    {
        uninstall(info);
    }
    else
    {
        /* If aborting an upgrade, just run the {pre,post}uninstall scripts */
        install_preuninstall(info);
        install_postuninstall(info);
    }
    exit_setup(0);
}

/* Abort a running installation (to be called from the update function) */
void abort_install(void)
{
    signal_abort(1);
}
    
/* List of UI drivers */
static int (*GUI_okay[])(Install_UI *UI) = {
    gtkui_okay,
    NULL
};

/* Matches a locale string against the current one */

int MatchLocale(const char *str)
{
	if ( ! str )
		return 1;
	if ( current_locale && ! (!strcmp(current_locale, "C") || !strcmp(current_locale,"POSIX")) ) {
		if ( strstr(current_locale, str) == current_locale ) {
			return 1;
		}
	} else if ( !strcmp(str, "none") ) {
		return 1;
	}
	return 0;
}

/* List the valid command-line options */

static void print_usage(const char *argv0)
{
    printf(
#ifdef RPM_SUPPORT
_("Usage: %s [options]\n\n"
"Options can be one or more of the following:\n"
"   -b path  Set the binary path to <path>\n"
"   -h       Display this help message\n"
"   -i path  Set the install path to <path>\n"
"   -c cwd   Use an alternate current directory for the install\n"
"   -f file  Use an alternative XML file (default %s)\n"
"   -m       Force manual extraction of RPM packages, do not update RPM database\n"
"   -n       Force the text-only user interface\n"
"   -o opt   Enable the option named \"opt\" from the XML file. Also enables non\n"
"            interactive operation. Can be used multiple times.\n"
"   -r root  Set the root directory for extracting RPM files (default is /)\n"
"   -v n     Set verbosity level to n. Available values :\n"
"            0: Debug  1: Quiet  2: Normal 3: Warnings 4: Fatal\n"
"   -V       Print the version of the setup program and exit\n"),
#else
_("Usage: %s [options]\n\n"
"Options can be one or more of the following:\n"
"   -b path  Set the binary path to <path>\n"
"   -h       Display this help message\n"
"   -i path  Set the install path to <path>\n"
"   -c cwd   Use an alternate current directory for the install\n"
"   -f file  Use an alternative XML file (default %s)\n"
"   -n       Force the text-only user interface\n"
"   -o opt   Enable the option named \"opt\" from the XML file. Also enables non\n"
"            interactive operation. Can be used multiple times.\n"
"   -v n     Set verbosity level to n. Available values :\n"
"            0: Debug  1: Quiet  2: Normal 3: Warnings 4: Fatal\n"
"   -V       Print the version of the setup program and exit\n"),
#endif
     argv0, SETUP_CONFIG);
}

/* Get a mount point for the specified CDROM, and return its path.
   If the CDROM is not mounted, prompt the user to mount it */
const char *get_cdrom(install_info *info, const char *id)
{
    const char *mounted = NULL, *desc = info->desc;
    struct cdrom_elem *cd;

    while ( ! mounted ) {
        for ( cd = info->cdroms_list; cd; cd = cd->next ) {
            if ( !strcmp(id, cd->id) ) {
                desc = cd->name;
                if ( cd->mounted ) {
                    mounted = cd->mounted;
                    break;
                }
            }
        }
        if ( ! mounted ) {
            yesno_answer response;
            char buf[1024];
            const char *prompt;

            if ( info->mounted_list ) { /* We were able to mount at least one CDROM */
                prompt =  _("\nPlease insert the %s CDROM.\n"
                            "Choose Yes to retry, No to cancel");                
            } else {
                prompt =  _("\nPlease mount the %s CDROM.\n"
                            "Choose Yes to retry, No to cancel");
            }
            unmount_filesystems(info);
            snprintf(buf, sizeof(buf), prompt, desc);
            response = UI.prompt(buf, RESPONSE_NO);
            if ( response == RESPONSE_NO ) {
                abort_install();
                return NULL;
            }
            detect_cdrom(info);
        }
    }
    return mounted;
}

/* Displays a dialog using the UI code and abort the installation */
void ui_fatal_error(const char *txt, ...)
{
    va_list ap;
    char buf[BUFSIZ];

    va_start(ap, txt);
    vsnprintf(buf, BUFSIZ, txt, ap);
    va_end(ap);

    UI.prompt(buf, RESPONSE_OK);
    abort_install();
}

/* The main installer code */
int main(int argc, char **argv)
{
    int exit_status;
    int i, c;
    install_state state;
    char *xml_file = SETUP_CONFIG;
    int log_level = LOG_NORMAL;
    char install_path[PATH_MAX];
    char binary_path[PATH_MAX];
    struct enabled_option *enabled_opt;

    install_path[0] = '\0';
    binary_path[0] = '\0';

    /* Set a good default umask value (022) */
    umask(DEFAULT_UMASK);

    /* 'Unset' SETUP_ABORT so the scripts are not confused */
    putenv("SETUP_ABORT=");

	/* Set the locale */
	setlocale (LC_ALL, "");
	bindtextdomain (PACKAGE, LOCALEDIR);
	textdomain (PACKAGE);

	current_locale = getenv("LC_ALL");
	if(!current_locale) {
		current_locale = getenv("LC_MESSAGES");
		if(!current_locale) {
			current_locale = getenv("LANG");
		}
	}

    /* Parse the command-line options */
    while ( (c=getopt(argc, argv,
#ifdef RPM_SUPPORT
					  "hnc:f:r:v:Vi:b:mo:"
#else
					  "hnc:f:v:Vi:b:o:"
#endif
					  )) != EOF ) {
        switch (c) {
		case 'c':
			if ( chdir(optarg) < 0 ) {
				perror(optarg);
				exit(-1);
			}
			break;
		case 'f':
			xml_file = optarg;
			break;
		case 'n':
			force_console = 1;
			break;
#ifdef RPM_SUPPORT
		case 'r':
			rpm_root = optarg;
			break;
#endif
		case 'v':
			if ( optarg ) {
				log_level = atoi(optarg);
				if ( (log_level < LOG_DEBUG) || (log_level > LOG_FATAL) ){
					fprintf(stderr,
							_("Out of range value, setting verbosity level to normal.\n"));
					log_level = LOG_NORMAL;
				}
			} else {
				log_level = LOG_DEBUG;
			}
			break;
		case 'V':
			printf("Loki Setup version " SETUP_VERSION ", built on "__DATE__"\n");
			exit(0);
	    case 'i':
	        strcpy(install_path, optarg);
			disable_install_path = 1;
 			break;
	    case 'b':
	        strcpy(binary_path, optarg);
			disable_binary_path = 1;
			break;
        case 'o': /* Store the enabled options for later processing */
            enabled_opt = (struct enabled_option *)malloc(sizeof(struct enabled_option));
            enabled_opt->option = strdup(optarg);
            enabled_opt->next = enabled_options;
            enabled_options = enabled_opt;
            break;
#ifdef RPM_SUPPORT
	    case 'm':
	        force_manual = 1;
			break;
#endif
		default:
			print_usage(argv[0]);
			exit(0);
        }
    }

	InitPlugins();
	if ( log_level == LOG_DEBUG ) {
		DumpPlugins(stderr);
	}

    /* Initialize the XML setup configuration */
    info = create_install(xml_file, log_level, install_path, binary_path);
    if ( info == NULL ) {
        fprintf(stderr, _("Couldn't load '%s'\n"), xml_file);
        exit(1);
    }

	log_debug(info, _("Detected locale is %s"), current_locale);

    /* Get the appropriate setup UI */
    for ( i=0; GUI_okay[i]; ++i ) {
        if ( GUI_okay[i](&UI) ) {
            break;
        }
    }
    if ( ! GUI_okay[i] ) {
        /* fprintf(stderr, _("No UI drivers available\n")); */
        exit(1);
    }

    /* Setup the interrupt handlers */
    state = SETUP_INIT;
    signal(SIGINT, signal_abort);
    signal(SIGHUP, signal_abort);
    signal(SIGTERM, signal_abort);

    /* Run the little state machine */
    exit_status = 0;
    while ( state != SETUP_EXIT ) {
        char buf[1024];
        int num_cds;
        char path[PATH_MAX];
	const char *warning=NULL;
	const char *disallow_install=NULL;
	int warn=0;

        switch (state) {
            case SETUP_INIT:
                state = UI.init(info,argc,argv, enabled_options != NULL);
                if ( state == SETUP_ABORT ) {
                    exit_status = 1;
                }

                /*------------------------------------------------------------------------
                ** Test for non root installations
                **----------------------------------------------------------------------*/
                if (GetProductNoWarnIfDirExists(info))
		{
                    struct stat dummy;
                    expand_home(info, GetProductNoWarnIfDirExists(info), path);
		    warn = stat(path, &dummy);
		}

		if (getuid())
		{
		    disallow_install = GetProductDisallowUserInstall(info);
		    if (GetProductDisallowRootInstall(info))
		        warn = 0;
		} else {
		    disallow_install = GetProductDisallowRootInstall(info);
		    if (GetProductDisallowUserInstall(info))
		        warn = 0;
		}

                /* Make sure $HOME belongs to us */
                if (disallow_install==NULL && getenv("HOME"))
                {
                    struct stat dummy;
                    int rc;

                    rc = stat(getenv("HOME"), &dummy);
                    if (rc < 0 || dummy.st_uid!=getuid())
                        disallow_install=_("'$HOME' must exist and belong to you in order for the installation to proceed. You may need to log in as root or use su rather than sudo.");
                }

		if (disallow_install)
		{
	            warn = 0;
		    UI.prompt(disallow_install, RESPONSE_OK);
		    state = SETUP_EXIT;
		}
			
		if (getuid() && warn)
                    warning = GetProductUserInstallWarning(info);
                if (!getuid() && warn)
                    warning = GetProductRootInstallWarning(info);
		
		if (warning)
		{
                    yesno_answer response;
                    response = UI.prompt(warning, RESPONSE_NO);
                    if (response == RESPONSE_NO)
                        state = SETUP_EXIT;
		}

                /* Check for the presence of the product if we install a component */
                if ( GetProductComponent(info) ) {
                    if ( GetProductNumComponents(info) > 0 ) {
                        UI.prompt(_("\nIllegal installation: do not mix components with a component installation.\n"), RESPONSE_OK);
                        state = SETUP_EXIT;
                    } else if ( info->product ) {
                        if ( ! info->component ) {
                            snprintf(buf, sizeof(buf), _("\nThe %s component is already installed.\n"
                                                         "Please uninstall it beforehand.\n"),
                                     GetProductComponent(info));
                            UI.prompt(buf, RESPONSE_OK);
                            state = SETUP_EXIT;
                        }
                    } else {                        
                        snprintf(buf, sizeof(buf), _("\nYou must install %s before running this\n"
													 "installation program.\n"),
                                info->desc);
                        UI.prompt(buf, RESPONSE_OK);
                        state = SETUP_EXIT;
                    }
                }

                num_cds = GetProductCDROMDescriptions(info);

                /* Check for the presence of a CDROM if required */
                if ( GetProductCDROMRequired(info) ) {
                    if ( ! GetProductCDROMFile(info) ) {
                        log_fatal(info, _("The 'cdromfile' attribute is now mandatory when using the 'cdrom' attribute."));
                    }
                    add_cdrom_entry(info, info->name, info->desc, GetProductCDROMFile(info));
                    detect_cdrom(info);

                    if ( ! get_cdrom(info, info->name) ) {
                        state = SETUP_EXIT;
                    }
                } else if ( num_cds > 0) {
                    detect_cdrom(info);
                }

                if ( enabled_options ) {
                    enabled_opt = enabled_options;
                    while ( enabled_opt ) {
                        if ( enable_option(info, enabled_opt->option) == 0 ) {
                            log_warning(info, _("Could not enable option: %s"), enabled_opt->option);
                        }
                        enabled_opt = enabled_opt->next;
                    }
                    state = SETUP_INSTALL;
                }
                break;
            case SETUP_LICENSE:
                state = UI.license(info);
                break;
            case SETUP_README:
                state = UI.readme(info);
                break;
            case SETUP_OPTIONS:
                state = UI.setup(info);
                break;
            case SETUP_INSTALL:
                if (install_preinstall(info) != 0) {state = SETUP_ABORT; break;};
                state = install(info, UI.update);
                install_postinstall(info);
                break;
            case SETUP_WEBSITE:
                state = UI.website(info);
                break;
            case SETUP_COMPLETE:
                state = UI.complete(info);
                break;
            case SETUP_PLAY:
                state = launch_game(info);
                break;
            case SETUP_ABORT:
                abort_install();
                break;
            case SETUP_EXIT:
                /* Not reached */
                break;
        }
    }

    /* Free enabled_options */
    while ( enabled_options ) {
        enabled_opt = enabled_options;
        enabled_options = enabled_options->next;
        free(enabled_opt->option);
        free(enabled_opt);
    }

    exit_setup(exit_status);
    return 0;
}
