/*
 * begining of a USB device handler
 *
 * Copyright 2004 Aric Stewart for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"
#include "wine/port.h"
#include "wine/library.h"

#include <stdarg.h>

#include "windef.h"
#include "winbase.h"
#include "winuser.h"
#include "wingdi.h"
#include "winnls.h"
#include "winreg.h"
#include "setupapi.h"
#include "wine/debug.h"
#include "winioctl.h"
#include "setupapi_private.h"
#include "wine/unicode.h"

#define INIT_GUID
#include "initguid.h"

#include "ntddstor.h"
#include "devguid.h"
#include "dbt.h"

/* from libusb */
#include "usb.h"
#include "usbdevice.h"

WINE_DEFAULT_DEBUG_CHANNEL(setupapi);

#define SONAME_USBNAME "libusb-0.1.4"
#define MAKE_FUNCPTR(f) static typeof(f) * p##f = NULL
MAKE_FUNCPTR(usb_init);
MAKE_FUNCPTR(usb_find_busses);
MAKE_FUNCPTR(usb_find_devices);
MAKE_FUNCPTR(usb_get_busses);
MAKE_FUNCPTR(usb_open);
MAKE_FUNCPTR(usb_get_string_simple);
MAKE_FUNCPTR(usb_close);
#undef MAKE_FUNCPTR

static void* libusb_handle = NULL;

struct usb_bus *busses;

static void USB_AddDevice(struct usb_device *dev, devinfo_struct *devinfo, 
                          const usb_device_map* map);

BOOL WINAPI USB_GetDeviceRegistryProperty(
        devnode *node,
        DWORD   Property,
        PDWORD  PropertyRegDataType,
        PBYTE   PropertyBuffer,
        DWORD   PropertyBufferSize,
        PDWORD  RequiredSize,
        BOOL    unicode);


static BOOL USB_SetupUSB()
{
    if (libusb_handle)
        return TRUE;

    libusb_handle = wine_dlopen(SONAME_USBNAME,RTLD_NOW, NULL, 0);
    if (!libusb_handle)
    {
        WINE_MESSAGE("Cannot load libusb. unable to process usb devices\n");
        return FALSE;
    }

#define LOAD_FUNCPTR(f) if((p##f = wine_dlsym(libusb_handle, #f, NULL, 0)) == NULL){WARN("Can't find symbol %s\n", #f);}

    LOAD_FUNCPTR(usb_init);
    LOAD_FUNCPTR(usb_find_busses);
    LOAD_FUNCPTR(usb_find_devices);
    LOAD_FUNCPTR(usb_get_busses);
    LOAD_FUNCPTR(usb_open);
    LOAD_FUNCPTR(usb_get_string_simple);
    LOAD_FUNCPTR(usb_close);

#undef LOAD_FUNCPTR

    pusb_init();
    pusb_find_busses();
    pusb_find_devices();
   
    busses = pusb_get_busses();

    return TRUE;
}

UINT USB_GetDeviceDevs(devinfo_struct *devinfo, CONST GUID *class, 
                       LPCSTR enumstr)
{
    int count = 0;
    struct usb_bus *bus;
    struct usb_device *dev;

    if (!USB_SetupUSB())
        return 0;

    TRACE("Looking for usb devices\n");

    for (bus = busses; bus; bus = bus->next)
    {
        for (dev = bus->devices; dev; dev = dev->next)
        {
            BOOL match = FALSE;
            INT i = 0, j=0, k=0, h=0;
            TRACE("found 0x%x / 0x%x\n",dev->descriptor.idVendor,
                  dev->descriptor.idProduct);
            {
            for (i = 0; i < dev->descriptor.bNumConfigurations && !match;
                 i++) 
            {
            for (j = 0; j < dev->config[i].bNumInterfaces && !match; j++)
            {
            for (k = 0; k < dev->config[i].interface[j].num_altsetting &&
                 !match; k++) 
            {
                struct usb_interface_descriptor *interface
                = &(dev->config[i].interface[j].altsetting[k]);
                h=0;
                if (devinfo->flags & DIGCF_DEVICEINTERFACE)
                {
                    const usb_device_map* entry = &known_msc_interfaces[0];
                    while (entry->class >=0 && !match)
                    {
                        if ((entry->class == interface->bInterfaceClass) &&
                            (entry->subclass ==
                             interface->bInterfaceSubClass) &&
                            (IsEqualCLSID(class,entry->devclass)))
                        {
                            match = TRUE;
                            USB_AddDevice(dev,devinfo,entry);
                            count ++;
                        }
                        h++;
                        entry = &known_msc_interfaces[h];
                    }
                }
                else if (devinfo->flags & DIGCF_ALLCLASSES)
                {
                    const usb_device_map* entry = &known_usb_devices[0];
                    while (entry->class >=0 && !match)
                    {
                        if ((entry->class == interface->bInterfaceClass) &&
                            (entry->subclass ==
                             interface->bInterfaceSubClass))
                        {
                            match = TRUE;
                            USB_AddDevice(dev,devinfo,entry);
                            count ++;
                        }
                        h++;
                        entry = &known_usb_devices[h];
                    }
                }
                else
                {
                    const usb_device_map* entry = &known_usb_devices[0];
                    while (entry->class >=0 && !match)
                    {
                        if ((entry->class == interface->bInterfaceClass) &&
                            (entry->subclass ==
                             interface->bInterfaceSubClass) &&
                            (IsEqualCLSID(class,entry->devclass)))
                        {
                            match = TRUE;
                            USB_AddDevice(dev,devinfo,entry);
                            count ++;
                        }
                        h++;
                        entry = &known_usb_devices[h];
                    }
                }
            }
            }
            }
            }
        }
    }
    TRACE("Found %i matching devices\n",count);
    return count;
}

static void USB_AddDevice(struct usb_device *dev, devinfo_struct *devinfo,
                          const usb_device_map* map)
{
    static const WCHAR diskW[] = {'d','i','s','k',0};
    static const WCHAR unknownW[] = {'u','n','k','n','o','w','n',0};
    static const WCHAR formatW[] = {'%','0','4','x',0};
    static const WCHAR revisionW[] = {'%','x','.','%','0','2','x',0};
    devnode* newdevices;
    int i,j;

    TRACE("Adding Device\n");
    i = devinfo->count;
    devinfo->count ++;
    if (devinfo->count > 1)
        newdevices = HeapReAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, 
                                 devinfo->devices,
                                 devinfo->count*sizeof(devnode));
    else
        newdevices = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
                               sizeof(devnode));


    /* faked for ipod for now to test structure */   
    memcpy(&(newdevices[i].guid),map->devclass,sizeof(GUID));
    newdevices[i].registry_fn = (device_registry_fn*)
                                &USB_GetDeviceRegistryProperty;
    newdevices[i].flags = 0x1;
    
    /* these need to be unicodified  in our datat structure*/
    MultiByteToWideChar(CP_ACP, 0, map->enumerator, -1,
                        newdevices[i].enumerator, 20);

    /* 
       all of these strings are presumed from 1 device... an ipod
       so they all may not be calculated correctly but at least i am 
       building a structure that can be revised as more information
       is gained
    */
    /* device type */
    switch (map->class)
    {
        case USB_CLASS_MASS_STORAGE:
            strcpyW(newdevices[i].devicetype,diskW);
            break;
        default:
            strcpyW(newdevices[i].devicetype,unknownW);
            break;
    }

    /* vendor  and product strings */
    /* 
        The question here is should i use usb.ids from linux-usb.org? 
        the issue is that is appears that windows translates these ides into
        VEN_XXXXX and PROD_XXXXXX strings that do not match the ids. 
        sometimes they are VEN_APPLE and other times it is VEN_05AC i am not
        sure when and how it does its translation. However it may also be
        the case that it pulls the string from the Manufacture string located
        on the device (like it pulls the serial number)  More investigation is
        needed.
    */

    j = 0;
    while (usb_vendor_product[j].id > 0)
    {
        if (usb_vendor_product[j].id == dev->descriptor.idVendor)
        {
            int k = 0;
            strcpyW(newdevices[i].vendor,usb_vendor_product[j].string);
            while ( usb_vendor_product[j].products[k].id > 0)
            {
                if (usb_vendor_product[j].products[k].id ==
                    dev->descriptor.idProduct)
                {
                    strcpyW(newdevices[i].product,
                            usb_vendor_product[j].products[k].string);
                    break;
                }
                k ++; 
            }
            break;
        }
        j++;
    }
    /* no string found cases */
    if (newdevices[i].vendor[0] == 0)
        sprintfW(newdevices[i].vendor,formatW,dev->descriptor.idVendor);
    if (newdevices[i].product[0] == 0)
        sprintfW(newdevices[i].product,formatW,dev->descriptor.idProduct);
    
    /* revision */
    /* here is the rub... linux usb is not! returning the revision correctly */
   
    if (dev->descriptor.bcdDevice)
    { 
        sprintfW(newdevices[i].revision, revisionW,
               dev->descriptor.bcdDevice>>8, dev->descriptor.bcdDevice & 0xff);
    }
    else
    {
        /* faked for the ipod... bad me */
        FIXME("Faking revision\n");
        sprintfW(newdevices[i].revision, revisionW, 1, 0x51);
    }

    /* serial number */
    /*
        This requires permissions to access and open the /proc/bus/usb/YYYY/XXXX
        file with RW privilages. The 'plan' is to make use of hotplug to make
        sure the permissions are set properly. (like gphoto does)
     */
    if (dev->descriptor.iSerialNumber)
    {
        CHAR device_string[256];
        usb_dev_handle *udev;
        INT ret;

        udev = pusb_open(dev);
        ret = pusb_get_string_simple(udev,dev->descriptor.iSerialNumber,
                                    device_string, 0x256); 
        if (ret > 0)
            MultiByteToWideChar(CP_ACP, 0, device_string, -1,
                                newdevices[i].serial, 50);
        else
            MESSAGE("Unable to fetch string: USB permissions incorrect?\n");

        pusb_close(udev);
    }
    newdevices[i].reserved = dev;

    devinfo->devices = newdevices;
}

/*
 * Example device id  from cfgmgr32
 *
 * "USBSTOR\\DISK&VEN_APPLE&PROD_IPOD&REV_1.51\\00000075261C&0\0"
 *
 */
static void USB_GetDeviceID(devnode* device, LPWSTR buffer, PDWORD size)
{
    static const WCHAR bashW[] = {'\\',0};
    static const WCHAR ampW[] = {'&',0};
    static const WCHAR prod_W[] = {'p','r','o','d','_',0};
    static const WCHAR ven_W[] = {'v','e','n','_',0};
    static const WCHAR rev_W[] = {'r','e','v','_',0};
    static const WCHAR zeroW[] = {'0',0};

    if (!device->deviceId[0])
    {
        /* build the device path */
        strcpyW(device->deviceId,device->enumerator);
        strcatW(device->deviceId,bashW);
        strcatW(device->deviceId,device->devicetype);
        strcatW(device->deviceId,ampW);
        strcatW(device->deviceId,ven_W);
        strcatW(device->deviceId,device->vendor);
        strcatW(device->deviceId,ampW);
        strcatW(device->deviceId,prod_W);
        strcatW(device->deviceId,device->product);
        strcatW(device->deviceId,ampW);
        strcatW(device->deviceId,rev_W);
        strcatW(device->deviceId,device->revision);
        strcatW(device->deviceId,bashW);
        if (device->serial[0])
        {
            strcatW(device->deviceId,device->serial);
            strcatW(device->deviceId,ampW);
            strcatW(device->deviceId,zeroW);
        }
        else
        {
            strcatW(device->deviceId,zeroW);
            strcatW(device->deviceId,ampW);
            strcatW(device->deviceId,zeroW);
        }
       
        TRACE("Built devid is %s\n",debugstr_w(device->deviceId));
    }

    if (buffer != NULL && *size >= strlenW(device->deviceId)+1)
        strcpyW(buffer,device->deviceId);

    *size = strlenW(device->deviceId) + 1;
    return;
};


static void USB_GetDevicePath(devnode* device, LPWSTR buffer, PDWORD size)
{
    static const WCHAR bashbashdotbashW[] = {'\\','\\','.','\\',0};
    static const WCHAR sharpW[] = {'#',0};
    static const WCHAR ampW[] = {'&',0};
    static const WCHAR prod_W[] = {'p','r','o','d','_',0};
    static const WCHAR ven_W[] = {'v','e','n','_',0};
    static const WCHAR rev_W[] = {'r','e','v','_',0};
    static const WCHAR zeroW[] = {'0',0};
    static const WCHAR guid_formatW[] =
{'{','%','0','8','l','x','-','%','0','4','x','-','%','0','4','x','-','%','0',
'2','x','%','0','2','x','-','%','0','2','x','%','0','2','x','%','0','2','x',
'%','0','2','x','%','0','2','x','%','0','2','x','}',0};
    WCHAR guid_string[40];

    if (!device->devicePath[0])
    {
        /* build the device path */
        strcpyW(device->devicePath,bashbashdotbashW);
        strcatW(device->devicePath,device->enumerator);
        strcatW(device->devicePath,sharpW);
        strcatW(device->devicePath,device->devicetype);
        strcatW(device->devicePath,ampW);
        strcatW(device->devicePath,ven_W);
        strcatW(device->devicePath,device->vendor);
        strcatW(device->devicePath,ampW);
        strcatW(device->devicePath,prod_W);
        strcatW(device->devicePath,device->product);
        strcatW(device->devicePath,ampW);
        strcatW(device->devicePath,rev_W);
        strcatW(device->devicePath,device->revision);
        strcatW(device->devicePath,sharpW);
        if (device->serial[0])
        {
            strcatW(device->devicePath,device->serial);
            strcatW(device->devicePath,ampW);
            strcatW(device->devicePath,zeroW);
        }
        else
        {
            strcatW(device->devicePath,zeroW);
            strcatW(device->devicePath,ampW);
            strcatW(device->devicePath,zeroW);
        }
       
        strcatW(device->devicePath,sharpW);
        sprintfW(guid_string, guid_formatW, device->guid.Data1,
                    device->guid.Data2, device->guid.Data3,
                    device->guid.Data4[0], device->guid.Data4[1],
                    device->guid.Data4[2], device->guid.Data4[3],
                    device->guid.Data4[4], device->guid.Data4[5],
                    device->guid.Data4[6], device->guid.Data4[7]
                 );
        strcatW(device->devicePath,guid_string);
    
        TRACE("Built devpath is %s\n",debugstr_w(device->devicePath));
    }

    if (buffer != NULL && *size >= strlenW(device->devicePath)+1)
        strcpyW(buffer,device->devicePath);

    *size = strlenW(device->devicePath) + 1;
    return;
};

static BOOL USB_GetHardwareId(devnode *node, LPWSTR buffer, PDWORD size)
{
    /* 
        Here is an example hardware id for an ipod under win2k

"USBSTOR\\DiskApple___iPod____________1.51\0USBSTOR\\DiskApple___iPod____________\0USBSTOR\\DiskApple___\0USBSTOR\\Apple___iPod____________1\0Apple___iPod____________1\0USBSTOR\\GenDisk\0GenDisk\0\0"

        enjoy
     */

     if (!node->hardwareid)
     {
        /*
         * build a very primitive hardware id
         */
        int len = 26;
        int i;
        static const WCHAR bashW[] = {'\\',0};
        static const WCHAR underscoreW[] = {'_',0};

        len += strlenW(node->enumerator);
        len += strlenW(node->devicetype);
        len += strlenW(node->revision);

        node->hardwareid = HeapAlloc(GetProcessHeap(),0,len * sizeof(WCHAR));

        strcpyW(node->hardwareid,node->enumerator);
        strcatW(node->hardwareid,bashW);
        strcatW(node->hardwareid,node->devicetype);
        strcatW(node->hardwareid,node->vendor);
        for (i = strlenW(node->vendor); i < 8; i++)
            strcatW(node->hardwareid,underscoreW);
        strcatW(node->hardwareid,node->product);
        for (i = strlenW(node->product); i < 16; i++)
            strcatW(node->hardwareid,underscoreW);
        strcatW(node->hardwareid,node->revision);

        TRACE("Built hardware id is %s\n",debugstr_w(node->hardwareid));
    }

    if (buffer != NULL && *size >= strlenW(node->hardwareid)+1)
        strcpyW(buffer,node->hardwareid);

    *size = strlenW(node->hardwareid) + 1;
    return TRUE;
}

BOOL WINAPI USB_GetDeviceRegistryProperty(
        devnode *node,
        DWORD   Property,
        PDWORD  PropertyRegDataType,
        PBYTE   PropertyBuffer,
        DWORD   PropertyBufferSize,
        PDWORD  RequiredSize,
        BOOL    unicode)
{
    /* 
        These things should be written to the CurrentConfiguration part 
        of the registry and just read out of there... but for now
        we can build them from out stored device data.
    */
    if (Property == SPDRP_HARDWAREID)
    {
        WCHAR buffer[PropertyBufferSize];
        DWORD size = PropertyBufferSize;

        USB_GetHardwareId(node, buffer, &size);
        if (RequiredSize)
        {
            *RequiredSize = size;
            if (unicode)
                *RequiredSize *= sizeof(WCHAR);
        }

        if (PropertyBufferSize >= size)
        {
            if (unicode)
                strcpyW((LPWSTR)PropertyBuffer, buffer);
            else
                WideCharToMultiByte(CP_ACP, 0, buffer, -1, PropertyBuffer,
                                    PropertyBufferSize, 0, 0);
        }
        return TRUE;
    }
    else if (Property == SPDRP_DEVICE_PATH)
    {
        WCHAR buffer[PropertyBufferSize];
        DWORD size = PropertyBufferSize;

        USB_GetDevicePath(node, buffer, &size);
        if (RequiredSize)
        {
            *RequiredSize = size;
            if (unicode)
                *RequiredSize *= sizeof(WCHAR);
        }

        if (PropertyBufferSize >= size)
        {
            if (unicode)
                strcpyW((LPWSTR)PropertyBuffer, buffer);
            else
                WideCharToMultiByte(CP_ACP, 0, buffer, -1, PropertyBuffer,
                                    PropertyBufferSize, 0, 0);
        }
        return TRUE;
    }
    else if (Property == SPDRP_DEVICE_ID)
    {
        WCHAR buffer[PropertyBufferSize];
        DWORD size = PropertyBufferSize;

        USB_GetDeviceID(node, buffer, &size);
        if (RequiredSize)
        {
            *RequiredSize = size;
            if (unicode)
                *RequiredSize *= sizeof(WCHAR);
        }

        if (PropertyBufferSize >= size)
        {
            if (unicode)
                strcpyW((LPWSTR)PropertyBuffer, buffer);
            else
                WideCharToMultiByte(CP_ACP, 0, buffer, -1, PropertyBuffer,
                                    PropertyBufferSize, 0, 0);
        }
        return TRUE;
    }

    return FALSE;
}

/*
http://msdn.microsoft.com/library/default.asp?url=/library/en-us/install/hh/install/idstrings_67a75d79-abf5-40bd-899e-097dd788fab4.xml.asp

usb device permissions
http://www.gphoto.org/doc/manual/permissions-usb.html
*/
